'use strict';

var path = require('path');

var Promise = require('bluebird');

var fs = Promise.promisifyAll(require('fs'));

var spawn = require('cross-spawn');

var debug = require('debug')('purs-loader');

var dargs = require('./dargs');

function compile(psModule) {
  var options = psModule.options;
  var cache = psModule.cache;
  var stderr = [];

  if (cache.compilationStarted) return Promise.resolve(psModule);

  cache.compilationStarted = true;

  var args = dargs(Object.assign({
    _: options.src,
    output: options.output
  }, options.pscArgs));

  debug('spawning compiler %s %o', options.psc, args);

  return new Promise(function (resolve, reject) {
    debug('\nCompiling PureScript...');

    var compilation = spawn(options.psc, args);

    compilation.stdout.on('data', function (data) {
      return stderr.push(data.toString());
    });
    compilation.stderr.on('data', function (data) {
      return stderr.push(data.toString());
    });

    compilation.on('close', function (code) {
      debug('Finished compiling PureScript.');
      cache.compilationFinished = true;
      if (code !== 0) {
        cache.errors = stderr.join('');
        reject(new Error('compilation failed'));
      } else {
        cache.warnings = stderr.join('');
        resolve(psModule);
      }
    });
  }).then(function (compilerOutput) {
    if (options.bundle) {
      return bundle(options, cache).then(function () {
        return psModule;
      });
    }
    return psModule;
  });
}
module.exports.compile = compile;

function bundle(options, cache) {
  if (cache.bundle) return Promise.resolve(cache.bundle);

  var stdout = [];
  var stderr = cache.bundle = [];

  var args = dargs(Object.assign({
    _: [path.join(options.output, '*', '*.js')],
    output: options.bundleOutput,
    namespace: options.bundleNamespace
  }, options.pscBundleArgs));

  cache.bundleModules.forEach(function (name) {
    return args.push('--module', name);
  });

  debug('spawning bundler %s %o', options.pscBundle, args.join(' '));

  return new Promise(function (resolve, reject) {
    debug('Bundling PureScript...');

    var compilation = spawn(options.pscBundle, args);

    compilation.stdout.on('data', function (data) {
      return stdout.push(data.toString());
    });
    compilation.stderr.on('data', function (data) {
      return stderr.push(data.toString());
    });
    compilation.on('close', function (code) {
      if (code !== 0) {
        cache.errors = (cache.errors || '') + stderr.join('');
        return reject(new Error('bundling failed'));
      }
      cache.bundle = stderr;
      resolve(fs.appendFileAsync(options.bundleOutput, 'module.exports = ' + options.bundleNamespace));
    });
  });
}